-- vim: ts=2 sw=2 ai et fenc=utf8
local MAJOR, MINOR = "LibGroupInSpecT-1.0", tonumber (("r22"):match ("(%d+)") or 0)

local lib = LibStub:NewLibrary (MAJOR, MINOR)
if not lib then return end

if lib.state then return end -- Already loaded

--[[
-- These events can be registered for using the regular CallbackHandler ways.
--
-- "GroupInSpecT_Update", guid, unit, info
-- "GroupInSpecT_Remove, guid
--
-- Where <info> is a table containing some or all of the following:
--   .guid
--   .name
--   .realm
--   .race
--   .race_localized
--   .class
--   .class_localized
--   .class_id
--   .gender -- 2 = male, 3 = female
--   .global_spec_id
--   .talents = {
--     [<tier>] = 0-3 (0 = none chosen)
--     ...
--   }
--   .glyphs = {
--     [glyphid] = GLYPHTYPE_MAJOR/GLYPTYPE_MINOR,
--     ...
--   },
--   .lku -- last known unit id
--
-- Functions for external use:
--
--   lib:Rescan ()
--     Force a rescan of all current group members
--
--   lib:QueuedInspections ()
--     Returns an array of GUIDs of outstanding inspects.
--
--   lib:StaleInspections ()
--     Returns an array of GUIDs for which the data has become stale and is
--     awaiting an update (no action required, the refresh happens internally).
--     Due to Blizzard exposing no events on (re/un)talent, there will be
--     frequent marking of inspect data as being stale.
--
--   lib:GetCachedInfo (guid)
--     Returns the cached info for the given GUID, if available, nil otherwise.
--     Information is cached for current group members only.
--
--   lib:GroupUnits ()
--     Returns an array with the set of unit ids for the current group.
--]]

local UPDATE_EVENT = "GroupInSpecT_Update"
local REMOVE_EVENT = "GroupInSpecT_Remove"

local INSPECT_DELAY = 1.5
local INSPECT_STALE_DELAY = 5
local INSPECT_TIMEOUT = 10 -- If we get no notification within 10s, give up on unit
local MAX_VALID_SPEC_ID = 500


local function debug (...)
--[===[@debug@
  --print (...)
--@end-debug@]===]
end


lib.events = LibStub ("CallbackHandler-1.0"):New (lib)


-- Frame for events
local frame = CreateFrame ("Frame", MAJOR .. "_Frame")
lib.frame = frame
frame:Hide()
frame:UnregisterAllEvents ()
frame:RegisterEvent ("PLAYER_ENTERING_WORLD")
frame:RegisterEvent ("PLAYER_LEAVING_WORLD")
frame:RegisterEvent ("INSPECT_READY")
frame:RegisterEvent ("GROUP_ROSTER_UPDATE")
frame:RegisterEvent ("UNIT_LEVEL")
frame:RegisterEvent ("PLAYER_TALENT_UPDATE")
frame:RegisterEvent ("PLAYER_SPECIALIZATION_CHANGED")
frame:RegisterEvent ("UNIT_SPELLCAST_SUCCEEDED")
frame:RegisterEvent ("UNIT_NAME_UPDATE")
frame:RegisterEvent ("GLYPH_ADDED")
frame:RegisterEvent ("GLYPH_REMOVED")
frame:RegisterEvent ("CHAT_MSG_ADDON")
frame:SetScript ("OnEvent", function (this, event, ...)
  return lib[event] (lib, ...)
end)

local COMMS_PREFIX = "LGIST1"
RegisterAddonMessagePrefix (COMMS_PREFIX)

-- Hide our run-state in an easy-to-dump object
lib.state = {
  mainq = {}, staleq = {}, -- inspect queues
  t = 0,
  last_inspect = 0,
  current_guid = nil,
  throttle = 0,
  tt = 0,
  debounce_send_update = 0,
}
lib.cache = {}


-- Get local handles on the key API functions
local CanInspect               = _G.CanInspect
local ClearInspectPlayer       = _G.ClearInspectPlayer
local GetGlyphSocketInfo       = _G.GetGlyphSocketInfo
local GetNumSubgroupMembers    = _G.GetNumSubgroupMembers
local GetPlayerInfoByGUID      = _G.GetPlayerInfoByGUID
local GetInspectSpecialization = _G.GetInspectSpecialization 
local GetSpecialization        = _G.GetSpecialization
local GetSpecializationInfo    = _G.GetSpecializationInfo 
local GetTalentInfo            = _G.GetTalentInfo
local IsInRaid                 = _G.IsInRaid
local NotifyInspect            = _G.NotifyInspect
local UnitClass                = _G.UnitClass
local UnitExists               = _G.UnitExists
local UnitGUID                 = _G.UnitGUID
local UnitInParty              = _G.UnitInParty
local UnitInRaid               = _G.UnitInRaid
local UnitIsConnected          = _G.UnitIsConnected
local UnitIsPlayer             = _G.UnitIsPlayer
local UnitIsUnit               = _G.UnitIsUnit
local UnitName                 = _G.UnitName
  

-- Inspects only work after entering the world, so track that
lib.state.logged_in = IsLoggedIn ()
function lib:PLAYER_ENTERING_WORLD ()
  lib.state.logged_in = true
  local guid = UnitGUID ("player")
  local info = self:BuildInfo ("player")
  self.events:Fire (UPDATE_EVENT, guid, "player", info)
end

function lib:PLAYER_LEAVING_WORLD ()
  lib.state.logged_in = false
end


-- Simple timer
do
  lib.state.t = 0
  frame:SetScript("OnUpdate", function(this, elapsed)
    lib.state.t = lib.state.t + elapsed
    lib.state.tt = lib.state.tt + elapsed
    local timeout = next (lib.state.mainq) and INSPECT_DELAY or INSPECT_STALE_DELAY
    if lib.state.t > timeout then
      lib:ProcessQueues ()
      lib.state.t = 0
    end
    -- Unthrottle, essentially allowing 1 msg every 3 seconds, but with substantial burst capacity
    if lib.state.tt > 3 and lib.state.throttle > 0 then
      lib.state.throttle = lib.state.throttle - 1
      lib.state.tt = 0
    end
    if lib.state.debounce_send_update > 0 then
      local debounce = lib.state.debounce_send_update - elapsed
      lib.state.debounce_send_update = debounce
      if debounce <= 0 then lib:SendLatestSpecData () end
    end
  end)
end


-- Internal library functions

local unitstrings = {
  raid = { "player" }, -- This seems to be needed under certain circumstances. Odd.
  party = { "player" }, -- Player not part of partyN
  player = { "player" }
}
for i = 1,40 do table.insert (unitstrings.raid, "raid"..i) end
for i = 1,4  do table.insert (unitstrings.party, "party"..i) end

-- Returns an array with the set of unit ids for the current group
function lib:GroupUnits ()
  local units
  if IsInRaid () then
    units = unitstrings.raid
  elseif GetNumSubgroupMembers () > 0 then
    units = unitstrings.party
  else
    units = unitstrings.player
  end
  return units
end


function lib:GuidToUnit (guid)
  local info = self.cache[guid]
  if info and info.lku and UnitGUID (info.lku) == guid then return info.lku end

  for i,unit in ipairs (self:GroupUnits ()) do
    if UnitExists (unit) and UnitGUID (unit) == guid then
      if info then info.lku = unit end
      return unit
    end
  end
end



function lib:Query (unit)
  if not UnitIsPlayer (unit) then return end -- NPC

  if UnitIsUnit (unit, "player") then
    self.events:Fire (UPDATE_EVENT, UnitGUID("player"), "player", self:BuildInfo ("player"))
    return
  end

  local mainq, staleq = self.state.mainq, self.state.staleq

  local guid = UnitGUID (unit)
  if not mainq[guid] then
    mainq[guid] = 1
    staleq[guid] = nil
    self.frame:Show () -- Start timer if not already running
  end
end


function lib:Refresh (unit)
  local guid = UnitGUID (unit)
  if not guid then return end
  debug ("Refreshing "..unit)
  if not self.state.mainq[guid] then
    self.state.staleq[guid] = 1
    self.frame:Show ()
  end
end


function lib:ProcessQueues ()
  if not self.state.logged_in then return end
  if InCombatLockdown () then return end -- Never inspect while in combat

  -- Don't' mess with the UI's inspections
	if InspectFrame and InspectFrame:IsShown () then
		return
	end

  local mainq = self.state.mainq
  local staleq = self.state.staleq

  local q = mainq
  if not next (q) then q = staleq end

  if (self.state.last_inspect + INSPECT_TIMEOUT) < GetTime () then
    -- If there was an inspect going, it's timed out so kick it out of the queue
    local guid = self.state.current_guid
    if guid then
      debug ("Inspect timed out for "..guid)
      if not self:GuidToUnit (guid) then
        debug ("No longer applicable, removing from queue")
        mainq[guid], staleq[guid] = nil, nil
      end
      self.state.current_guid = nil
    end
  end

  if self.state.current_guid then return end -- Still waiting on our inspect data

	for guid,_ in pairs (q) do
		local unit = self:GuidToUnit (guid)
		if not unit or not CanInspect (unit) or not UnitIsConnected (unit) then
			q[guid] = nil
		else
      debug ("Inspecting "..unit..", aka ".. (UnitName(unit) or "nil"))
      self.state.current_guid = guid
		  NotifyInspect (unit)
      break
    end
	end

  if not next (self.state.mainq) and not next (self.state.staleq) and self.state.throttle == 0 and self.state.debounce_send_update <= 0 then
    frame:Hide() -- Cancel timer, nothing queued and no unthrottling to be done
  end
end


hooksecurefunc("NotifyInspect", function (...) return lib:NotifyInspect (...) end)
function lib:NotifyInspect(unit)
	self.state.last_inspect = GetTime()
end

function lib:UpdatePlayerInfo (guid, unit, info)
  info.class_localized, info.class, info.race_localized, info.race, info.gender, info.name, info.realm = GetPlayerInfoByGUID (guid)
  if info.realm and info.realm == "" then info.realm = nil end
  local _, _, class_id = UnitClass (unit)
  info.class_id = class_id
  info.lku = unit
end

local function erase (table)
  if not table then return {} end
  for i,_ in pairs (table) do
    table[i] = nil
  end
  return table
end

function lib:BuildInfo (unit)
  local guid = UnitGUID (unit)
  if not guid then return end

  local cache = self.cache
  local info = cache[guid] or {}
  cache[guid] = info
  info.guid = guid

  self:UpdatePlayerInfo (guid, unit, info)

  local is_inspect = not UnitIsUnit("player", unit)
  local spec = GetSpecialization ()
  info.global_spec_id = is_inspect and GetInspectSpecialization (unit) or spec and GetSpecializationInfo (spec)
  if info.global_spec_id and info.global_spec_id > MAX_VALID_SPEC_ID then
    info.global_spec_id = nil
  end

  info.talents = erase (info.talents)
  local idx = 1
  local tname, tex, tier, column, sel, avail = GetTalentInfo (idx, is_inspect, nil, unit, info.class_id)
  while tier and column do
    if not info.talents[tier] then info.talents[tier] = 0 end
    if info.talents[tier] == 0 and sel then
      info.talents[tier] = column
    end
    idx = idx + 1
    tname, tex, tier, column, sel, avail = GetTalentInfo (idx, is_inspect, nil, unit, info.class_id)
  end
  info.glyphs = erase (info.glyphs)
  for i = 1,NUM_GLYPH_SLOTS do
    local _, glyphtype, _, glyphid = GetGlyphSocketInfo (i, nil, is_inspect, unit)
    if glyphid then info.glyphs[glyphid] = glyphtype end
  end
  return info
end

function lib:INSPECT_READY (guid)
	local unit = self:GuidToUnit (guid)
  local finalize = false
	if unit then
    if guid == self.state.current_guid then
      self.state.current_guid = nil -- Got what we asked for
      finalize = true
      debug("Got inspection data for requested guid "..guid)
    end

    local mainq, staleq = self.state.mainq, self.state.staleq
    mainq[guid], staleq[guid] = nil, nil

    local gspec_id = GetInspectSpecialization (unit)
    if gspec_id > MAX_VALID_SPEC_ID then -- Bah, got garbage, flag as stale and try again
      self.state.staleq[guid] = guid
      return
    end

    self.events:Fire (UPDATE_EVENT, guid, unit, self:BuildInfo (unit))
	end
  if finalize then
    ClearInspectPlayer ()
  end
end


-- Group handling parts

function lib:GROUP_ROSTER_UPDATE ()
  local group = self.cache
  local members = {}
  local units = self:GroupUnits ()
  -- Find new members
  for i,unit in ipairs (self:GroupUnits ()) do
    local guid = UnitGUID (unit)
    if guid then
      members[guid] = 1
      if not group[guid] then
        self:Query (unit)
        -- Update with what we have so far (guid, unit, name/class/race?)
        self.events:Fire (UPDATE_EVENT, guid, unit, self:BuildInfo (unit))
      end
    end
  end
  -- Find removed members
  for guid,info in pairs (group) do
    if not members[guid] then
      group[guid] = nil
      self.events:Fire (REMOVE_EVENT, guid, nil)
    end
  end
end

function lib:DoPlayerUpdate ()
  self:Query ("player")
  self.state.debounce_send_update = 2.5 -- Hold off 2.5sec before sending update
  frame:Show ()
end

function lib:SendLatestSpecData ()
  local guid = UnitGUID ("player")
  local info = self.cache[guid]
  if not info then return end
  -- "UPD,spec,talents,glyph=n:glyph=n..."
  local datastr = "UPD,"..guid..","..(info.global_spec_id or 0)..","
  for _,v in ipairs (info.talents) do
    datastr = datastr..v
  end
  datastr = datastr..","
  for i,t in pairs (info.glyphs) do
    datastr = datastr..i.."="..t..":"
  end

  if UnitInBattleground ("player") ~= nil then
    debug("Sending LGIST update to BATTLEGROUND: "..datastr)
    SendAddonMessage (COMMS_PREFIX, datastr, "BATTLEGROUND")
  elseif IsInRaid () or GetNumGroupMembers () > 0 then
    debug("Sending LGIST update to RAID: "..datastr)
    SendAddonMessage (COMMS_PREFIX, datastr, "RAID")
  end
end

function lib:CHAT_MSG_ADDON (prefix, datastr, scope, sender)
  if prefix ~= COMMS_PREFIX then return end
  debug ("Incoming LGIST update from "..(scope or "nil").."/"..(sender or "nil")..": "..(datastr or "nil"))

  self.state.throttle = self.state.throttle + 1
  frame:Show () -- Ensure we're unthrottling
  if self.state.throttle > 40 then return end -- If we ever hit this, someone's being "funny"

  local unit, guid, talents, glyphs, info
  if datastr:find ("UPD,") then
    guid, gsid, talents, glyphs = datastr:match ("UPD,(0x%x+),(%d+),(%d*),(.*)")
    info = guid and self.cache[guid]
    -- Only need to handle this message if we have a cached entry, else inspect will give full info anyway
    if not info then return end

    unit = self:GuidToUnit (guid)
    if unit == "player" then return end -- We're already up-to-date!

    info.global_spec_id = tonumber (gsid)
    info.talents = erase (info.talents)
    for i = 1,talents:len () do
      info.talents[i] = tonumber (talents:sub (i,i))
    end
    info.glyphs = erase (info.glyphs)
    for glyphinfo in glyphs:gmatch ("(%d+=%d):") do
      local glyphid,glyphtype = glyphinfo:match ("(%d+)=(%d)")
      info.glyphs[tonumber (glyphid)] = tonumber (glyphtype)
    end
  end

  if unit and guid then
    --debug("Firing LGIST update event for unit "..(unit or "nil")..", GUID "..guid)
    self.events:Fire (UPDATE_EVENT, guid, unit, info)
  end
end

function lib:UNIT_LEVEL (unit)
  if UnitInRaid (unit) or UnitInParty (unit) then
    self:Refresh (unit)
  end
  if UnitIsUnit (unit, "player") then
    self:DoPlayerUpdate ()
  end
end

function lib:PLAYER_TALENT_UPDATE ()
  self:DoPlayerUpdate ()
end

function lib:PLAYER_SPECIALIZATION_CHANGED (unit)
--  This event seems to fire a lot, and for no particular reason *sigh*
--  if UnitInRaid (unit) or UnitInParty (unit) then
--    self:Refresh (unit)
--  end
  if unit and UnitIsUnit (unit, "player") then
    self:DoPlayerUpdate ()
  end
end

function lib:GLYPH_ADDED ()
  self:DoPlayerUpdate ()
end

function lib:GLYPH_REMOVED ()
  self:DoPlayerUpdate ()
end

function lib:UNIT_NAME_UPDATE (unit)
  local group = self.cache
  local guid = UnitGUID (unit)
  local info = guid and group[guid]
  if info then
    self:UpdatePlayerInfo (guid, unit, info)
    if info.name ~= UNKNOWN then
      self.events:Fire (UPDATE_EVENT, guid, unit, info)
    end
  end
end

local dual_spec_spells = {}
for i, spellid in ipairs (TALENT_ACTIVATION_SPELLS) do dual_spec_spells[spellid] = true end
function lib:UNIT_SPELLCAST_SUCCEEDED (unit, spellname, rank, lineid, spellid)
  if dual_spec_spells[spellid] then
    self:Query (unit) -- Definitely changed, so high prio refresh
  end
end


-- External library functions

function lib:QueuedInspections ()
  local q = {}
  for guid,_ in pairs (self.state.mainq) do
    table.insert (q, guid)
  end
  return q
end

function lib:StaleInspections ()
  local q = {}
  for guid,_ in pairs (self.state.staleq) do
    table.insert (q, guid)
  end
  return q
end

function lib:GetCachedInfo (guid)
  local group = self.cache
  return guid and group[guid]
end

function lib:Rescan ()
  local q = self.state.mainq
  for i,unit in ipairs (self:GroupUnits ()) do
    if UnitExists (unit) then
      local guid = UnitGUID (unit)
      if guid then q[guid] = 1 end
    end
  end
  self.frame:Show () -- Start timer if not already running

  -- Evict any stale entries
  self:GROUP_ROSTER_UPDATE ()
end
